/* $Workfile: 2T2p.c $       $Revision: 43 $      $Modtime: 12/02/03 11:12a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    2T2p.c

    An example console application demonstrating the use of OTX Driver
    with n*2 
    
      Thor-2-PCI or 
      Thor-8-PCI or 
      Thor-8-PCI-Plus 
      
    boards connected over the CT Bus 
    (n can be specified with the /N command line switch).

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    Handsets are through-connected over the H.100 bus between two boards.

    Press 'h' or '?' for help. 'q' to quit.

    Creating a logical board interconnect device to represent a physical connection
    between two or more boards.

    When two ore more boards are connected together with the appropriate board
    CT Bus interconnect ribbon cable it is possible to connect two devices with a pipe
    even if they are located on different board. (This example support multiple CT Buses
    with two Thor-2-PCI boards connected to each CT Bus interconnect cable.)

    Before pipes can connect over separate board the driver must be informed about
    what types of board interconnects exists, and how the boards are connected.
    This is done by opening the physical device OTX_DEVICE_CT_BUS (which is hosted by
    the driver device (0)). The physical device OTX_DEVICE_CT_BUS is a representation
    of the CT Bus cable which connects two or more boards. OtxDrvConnectBoard() is called
    to inform the OTX_DEVICE_CT_BUS device about which boards that is connected to it.

    The CT Bus does not have a direction like a pipe. The direction of each individual
    timeslot is determined at the time a Pipe Device must utilize the CT Bus
    to reach a different board. The direction of the pipe then determines the
    direction of that timeslot in the CT Bus.

    The CT Bus device has several highways (the actual leads of the cable). Each highway 
    is then multiplexed into several logical timeslots. One timeslot is always 64kbit/s 
    so the total number of timeslots per highway is then determined by the capacity of 
    each highway. The standard capacity is 2048 kbits/s which corresponds to 32 ts/hwy.


*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if defined(_WIN32_WCE)
#include "Lib/WindowsCe/OsaWindowsCe.h"
#else   // #if defined(_WIN32_WCE)
#include <assert.h>
#endif  // #if defined(_WIN32_WCE)

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"

#define DEMO_DEFAULT_LOG_FILE "2T2p.log"
#define DEMO_NO_OF_LIS 2
#define DEMO_MAX_BOARDS 32

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call

// This example program uses a fixed connection method by opening the Time-Space Switch (TSS)
// physical device. Pipes connected to the TSS can be viewed as a pipe within the TSS device
// itself. The TSS device has 16 local highways, and 32 H.100 highways. To make a connection
// between a device (e.g. Codec) on a local highway and one H.100 highway, a pipe is first
// created, then connected using the handle to the TSS physical device.
// 
// Future releases of the driver will have a simplified way of connecting between
// two OTX boards where the actual highway number and timeslots used by the TSS will not
// have to be known by the application writer.
#define DEMO_H100_HWY_0      16
#define DEMO_H100_HWY_1      17
// ....
#define DEMO_H100_HWY_31     47

typedef struct DemoEnvT {
  char            *pszInputStr;
  OTX_BOOL        bExitOnError;
  char            *logFileName;
  OtxPhysicalDeviceTypeE nBoardType;
  OtxDataFormatE  law;
  OtxT1E1LiModeE  eT1E1Mode;    // T1 or E1 mode
  OTX_BOOL        bSs7FisuFiltering;
  OTX_INT32       coCtBuses;
  OTX_INT32       coBrds;
  OTX_INT32       nHbusRate;
  OtxBrdClockSourceE  eClkSource;
  OtxBrdClockModeE  eClkMode;
  OTX_UINT32      nHighwayCodec;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(OTX_INT32 iBrd);
static void DemoConfigureE1T1Devices(OTX_INT32 iBrd);
static void DemoCreateLogicalDevices(OTX_INT32 iBrd);
static void DemoConnectDevices(OTX_INT32 iBrd);
static void DemoEnableDevices(OTX_INT32 iBrd);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_HANDLE  g_hCtBus[16];

struct {
  OTX_HANDLE hThorController;
  OTX_HANDLE hQuadE1T1[2]; // Only used for Thor-8-PCI and Thor-8-PCI-Plus)
  OTX_HANDLE hTss;
  OTX_HANDLE hLi[2];
  OTX_HANDLE hCodec[4];
} g_hPhysDevs[DEMO_MAX_BOARDS];

struct {
  OTX_BOOL bOccupied[32];  // 32 timeslots per highway
} g_sH100Hwy[32]; // 32 H.100 highways

struct {
  OTX_INT32 nH100Hwy; 
  OTX_INT32 nTs0To1; // H100 timeslot use from board 0 to board 1;
  OTX_INT32 nTs1To0; // H100 timeslot use from board 1 to board 0;
  OTX_BOOL bConnected; 
} g_sHs[4]; 

struct {
  OTX_HANDLE hLiIn[2];
  OTX_HANDLE hLiOut[2];
  OTX_HANDLE hHsMicPipe[4];
  OTX_HANDLE hHsSpkrPipe[4];
} g_hPipes[DEMO_MAX_BOARDS];

struct {
  OTX_HANDLE hHdlcSender[2];
  OTX_HANDLE hHdlcReceiver[2];
} g_hLogicDevs[DEMO_MAX_BOARDS];

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;
                               
  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects  
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor-2-PCI board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(OTX_INT32 iBrd)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening OTX_DEVICE_THOR_PCI board #%d\n", iBrd);
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board #%d\n", iBrd);
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board #%d\n", iBrd);
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, iBrd, iBrd, g_hEventQueue, &g_hPhysDevs[iBrd].hThorController); DemoCheckRc(g_hPhysDevs[iBrd].hThorController, result);

  // Get a handle to the Time-Space Switch (TSS) on the Thor-2-PCI board
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[iBrd].hThorController, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs[iBrd].hTss); DemoCheckRc(g_hPhysDevs[iBrd].hThorController, result);

  // Set the H.100 Rate (default is 2048 kbps)    
  DemoLogf("Setting rate to %d for all H.100 highways.\n", g_demoEnv.nHbusRate);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[iBrd].hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_A, g_demoEnv.nHbusRate); DemoCheckRc(g_hPhysDevs[iBrd].hTss, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[iBrd].hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_B, g_demoEnv.nHbusRate); DemoCheckRc(g_hPhysDevs[iBrd].hTss, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[iBrd].hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_C, g_demoEnv.nHbusRate); DemoCheckRc(g_hPhysDevs[iBrd].hTss, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[iBrd].hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_D, g_demoEnv.nHbusRate); DemoCheckRc(g_hPhysDevs[iBrd].hTss, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[iBrd].hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_E, g_demoEnv.nHbusRate); DemoCheckRc(g_hPhysDevs[iBrd].hTss, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[iBrd].hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_F, g_demoEnv.nHbusRate); DemoCheckRc(g_hPhysDevs[iBrd].hTss, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[iBrd].hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_G, g_demoEnv.nHbusRate); DemoCheckRc(g_hPhysDevs[iBrd].hTss, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[iBrd].hTss, OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_H, g_demoEnv.nHbusRate); DemoCheckRc(g_hPhysDevs[iBrd].hTss, result);
      
  DemoLogf("Enable OTX_DEVICE_TSS\n");
  result = OtxDrvEnable(g_hPhysDevs[iBrd].hTss, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs[iBrd].hTss, result);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    // On the Thor-2-PCI board the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < DEMO_NO_OF_LIS; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[iBrd].hThorController, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs[iBrd].hLi[i]); DemoCheckRc(g_hPhysDevs[iBrd].hThorController, result);
    }
    break;
  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    {
      OTX_INT32 nQuadE1T1; // Only used for Thor-8-PCI and Thor-8-PCI-Plus

      // On the Thor-8-PCI and Thor-8-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[iBrd].hThorController, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs[iBrd].hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs[iBrd].hThorController, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < DEMO_NO_OF_LIS; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[iBrd].hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs[iBrd].hLi[i]); DemoCheckRc(g_hPhysDevs[iBrd].hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }


  // Get handles to Codec devices #0 and #1
  for (i = 0; i < 4; i++) {
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[iBrd].hThorController, OTX_DEVICE_CODEC, i, i, g_hEventQueue, &g_hPhysDevs[iBrd].hCodec[i]); DemoCheckRc(g_hPhysDevs[iBrd].hThorController, result);
  }

}





/*******************************************************************************

  @func
    Configure E1/T1 Devices

*******************************************************************************/
void DemoConfigureE1T1Devices(OTX_INT32 iBrd)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Configure Transceivers
  for (i = 0; i < DEMO_NO_OF_LIS; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs[iBrd].hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs[iBrd].hLi[i], result);
  
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs[iBrd].hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs[iBrd].hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs[iBrd].hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs[iBrd].hLi[i], result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs[iBrd].hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs[iBrd].hLi[i], result);
  }

}




/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  OTX_INT32 iBrd  
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < DEMO_NO_OF_LIS; i++) {
    // Create Hdlc Senders and Receivers logical devices hosted in the physical Line Interface devices
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs[iBrd].hLi[i], OTX_LDEVICE_HDLC_SENDER, g_hEventQueue, iBrd * DEMO_NO_OF_LIS + i, &g_hLogicDevs[iBrd].hHdlcSender[i]); DemoCheckRc(g_hPhysDevs[iBrd].hLi[i], result);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs[iBrd].hLi[i], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, iBrd * DEMO_NO_OF_LIS + i, &g_hLogicDevs[iBrd].hHdlcReceiver[i]); DemoCheckRc(g_hPhysDevs[iBrd].hLi[i], result);
  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  OTX_INT32 iBrd  
)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 ts;

  for (i = 0; i < 4; i++) {
  // Create voice pipes for the 4 Handsets
    result = OtxDrvCreatePipe(64, &g_hPipes[iBrd].hHsMicPipe[i]); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[i], result);
    result = OtxDrvCreatePipe(64, &g_hPipes[iBrd].hHsSpkrPipe[i]); DemoCheckRc(g_hPipes[iBrd].hHsSpkrPipe[i], result);
  }

#if 0
  // Connect Hs0 <--> Hs1 in the time-space switch
  result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsMicPipe[0], OTX_PIPE_INPUT,  g_hPhysDevs[iBrd].hCodec[0], -1, NULL, 0); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[0], result);
  result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsMicPipe[0], OTX_PIPE_OUTPUT, g_hPhysDevs[iBrd].hCodec[1], -1, NULL, 0); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[0], result);
  
  result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsMicPipe[1], OTX_PIPE_INPUT,  g_hPhysDevs[iBrd].hCodec[1], -1, NULL, 0); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[1], result);
  result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsMicPipe[1], OTX_PIPE_OUTPUT, g_hPhysDevs[iBrd].hCodec[0], -1, NULL, 0); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[1], result);
#endif


  for (i = 0; i < DEMO_NO_OF_LIS; i++) {
    // Create an incoming and outgoing Pipe representing the external T1/E1
    result = OtxDrvCreatePipe(64, &g_hPipes[iBrd].hLiIn[i]); DemoCheckRc(g_hPipes[iBrd].hLiIn[i], result);
    result = OtxDrvCreatePipe(64, &g_hPipes[iBrd].hLiOut[i]); DemoCheckRc(g_hPipes[iBrd].hLiOut[i], result);

    // Connect external T1/E1 pipes. Only connect one end of each pipe. External Interface is LI Highway Number 1
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      ts = 24; 
    } else {
      // E1 mode
      ts = 16; 
    }
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[ts] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes[iBrd].hLiIn[i], OTX_PIPE_OUTPUT, g_hPhysDevs[iBrd].hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[iBrd].hLiIn[i], result);
    result = OtxDrvConnectPipe(g_hPipes[iBrd].hLiOut[i], OTX_PIPE_INPUT, g_hPhysDevs[iBrd].hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[iBrd].hLiOut[i], result);

    // Connect HDLC Sender and Receiver to the external T1/E1 Pipes
    result = OtxDrvConnectLogicalDevice(g_hPipes[iBrd].hLiIn[i], -1, g_hLogicDevs[iBrd].hHdlcReceiver[i], -1); DemoCheckRc(g_hPipes[iBrd].hLiIn[i], result);
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs[iBrd].hHdlcSender[i], -1, g_hPipes[iBrd].hLiOut[i], -1); DemoCheckRc(g_hLogicDevs[iBrd].hHdlcSender[i], result);

  }



}



/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  OTX_INT32 iBrd  
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < DEMO_NO_OF_LIS; i++) {

    // Initialize the Hdlc Sender and Receiver
    result = OtxHdlcSenderInit(
               g_hLogicDevs[iBrd].hHdlcSender[i], // Handle to logical HDLC Filter Device
               OTX_HDLC_FFT_FLAGS,          // Fill character(s) to be send between HDLC frames
               2,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
               OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
               OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
               OTX_TRUE,                    // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
               OTX_HDLC_TR_NA);             // HDLC Transmission Rate for V.110 and X.30 modes
               DemoCheckRc(g_hLogicDevs[iBrd].hHdlcSender[i], result);

    result = OtxHdlcReceiverInit(
               g_hLogicDevs[iBrd].hHdlcReceiver[i], // Handle to logical HDLC Filter Device
               OTX_HDLC_FFT_ALL_ONES,          // Fill character(s) to be send between HDLC frames
               1,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
               OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
               OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
               OTX_TRUE,                    // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
               OTX_HDLC_TR_NA,              // HDLC Transmission Rate for V.110 and X.30 modes
               OTX_TRUE);                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
               DemoCheckRc(g_hLogicDevs[iBrd].hHdlcReceiver[i], result);

    if (g_demoEnv.bSs7FisuFiltering) {
      result = OtxHdlcSS7SetFilter(g_hLogicDevs[iBrd].hHdlcReceiver[i], OTX_HDLC_SS7_FILTER_FISU); DemoCheckRc(g_hLogicDevs[iBrd].hHdlcReceiver[i], result);
    }
  }

}



/*******************************************************************************

  @func
    Enable Configured Devices

*******************************************************************************/
void DemoEnableDevices(OTX_INT32 iBrd)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Enable Physical Devices
  result = OtxDrvEnable(g_hPhysDevs[iBrd].hThorController,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs[iBrd].hThorController, result);

  // If board type is Thor-8-PCI and Thor-8-PCI-Plus then enable Quad E1T1 device
  if ((g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI) || (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS)) {
    OTX_INT32 nQuadE1T1; // Only used for Thor-8-PCI and Thor-8-PCI-Plus

    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs[iBrd].hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs[iBrd].hQuadE1T1[nQuadE1T1], result);
    }
  }

  for (i = 0; i < DEMO_NO_OF_LIS; i++) {
    result = OtxDrvEnable(g_hPhysDevs[iBrd].hLi[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs[iBrd].hLi[i], result);
  }

  // Enable Logical HDLC Devices (Data Sinks)
  for (i = 0; i < DEMO_NO_OF_LIS; i++) {
    result = OtxDrvEnable(g_hLogicDevs[iBrd].hHdlcSender[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs[iBrd].hHdlcSender[i], result);
    result = OtxDrvEnable(g_hLogicDevs[iBrd].hHdlcReceiver[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs[iBrd].hHdlcReceiver[i], result);
  }


}




/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_TRUE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 128;
  pE1CfgData->nLosRecovery               = 128;
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 128;
  pT1CfgData->nLosRecovery               = 128;
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x00;
}





/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;

}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  OTX_UINT8  buf[1024];
  OTX_UINT32 nFrameLen;
  OTX_UINT32 nFrameSeq;
  OtxHdlcFrameEndCodeE nFrameEndCode;
  OTX_DATETIME nTimeStamp;
  char msgStr[256];
    
  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_TSS:
    // Event from the Time-Space Switch
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("TSS Event: %s \n", msgStr);
    } else {
      DemoLogf("TSS Invalid Event: %ld \n", pEventData->m_nCode);
    }
    break;
    
  case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;
    
  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HLDC Receiver 
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {

      otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);

      DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);

      if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
        DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
        for (i = 0; i < nFrameLen && i < 1024; i++) {
          DemoLogf("%.2X ", buf[i]);
        }
        DemoLogf("\n");
      } else {
        DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(pEventData->m_hDevice, otxRc);
      }

    } else {
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
    
  case OTX_LDEVICE_HDLC_SENDER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    break;
    
  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}



/*******************************************************************************

  @func

*******************************************************************************/
OTX_BOOL DemoConnectHsViaH100(
  OTX_INT32 nHs,
  OTX_INT32 nH100Hwy,
  OTX_INT32 nTs0To1,
  OTX_INT32 nTs1To0,
  OTX_BOOL bVerbose
)
{
  
  OTX_UINT8 sqTimeSlotMask[128];
  OTX_INT32 iBrd;
  OTX_INT32 nHsTs;
  OTX_INT32 nH100Ts;
  OTX_RESULT result;
  OTX_INT32 nMaxHbusTimeSlot = g_demoEnv.nHbusRate/64;

  if (nTs0To1 == nTs1To0) {
    DemoLogf("Error: nTs0To1 and nTs1To0 are identical. H100 highways are simplex. Must use different timeslot for each direction. Aborting this operation. \n", nTs0To1);
    return OTX_FALSE;
  }
  if (g_sH100Hwy[nH100Hwy].bOccupied[nTs0To1]) {
    DemoLogf("Error: Ts#%d (nTs0To1) is already occupied. Aborting this operation. \n", nTs0To1);
    return OTX_FALSE;
  }
  if (g_sH100Hwy[nH100Hwy].bOccupied[nTs1To0]) {
    DemoLogf("Error: Ts#%d (nTs1To0) is already occupied. Aborting this operation. \n", nTs1To0);
    return OTX_FALSE;
  }
  g_sH100Hwy[nH100Hwy].bOccupied[nTs0To1] = OTX_TRUE;
  g_sH100Hwy[nH100Hwy].bOccupied[nTs1To0] = OTX_TRUE;

  g_sHs[nHs].nTs0To1 = nTs0To1;
  g_sHs[nHs].nTs1To0 = nTs1To0;
  g_sHs[nHs].nH100Hwy = nH100Hwy;
  

  nHsTs = nHs + 1; // HS0 is TS1, HS1 is TS2, HS2 is TS3, HS3 is TS4 on the g_demoEnv.nHighwayCodec

  for (iBrd = 0; iBrd < g_demoEnv.coBrds; iBrd++) {
    if (bVerbose) DemoLogf("Board#%d: Connecting HS%d to H100 Hwy0, ", iBrd, nHs);
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[nHsTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsMicPipe[nHs], OTX_PIPE_INPUT,   g_hPhysDevs[iBrd].hTss, g_demoEnv.nHighwayCodec, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[nHs], result);
    
    if (iBrd == 0) {
      nH100Ts = nTs0To1;
    } else {
      nH100Ts = nTs1To0;
    }
    if (bVerbose) DemoLogf("Mic to TS%d, ", nH100Ts);
    memset(sqTimeSlotMask, 0, nMaxHbusTimeSlot * sizeof(OTX_UINT8));
    sqTimeSlotMask[nH100Ts] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsMicPipe[nHs], OTX_PIPE_OUTPUT,   g_hPhysDevs[iBrd].hTss, DEMO_H100_HWY_0 + nH100Hwy, sqTimeSlotMask, nMaxHbusTimeSlot); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[nHs], result);
    
    if (iBrd == 0) {
      nH100Ts = nTs1To0;
    } else {
      nH100Ts = nTs0To1;
    }
    if (bVerbose) DemoLogf("Spkr to TS%d \n", nH100Ts);
    memset(sqTimeSlotMask, 0, nMaxHbusTimeSlot * sizeof(OTX_UINT8));
    sqTimeSlotMask[nH100Ts] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsSpkrPipe[nHs], OTX_PIPE_INPUT,   g_hPhysDevs[iBrd].hTss, DEMO_H100_HWY_0 + nH100Hwy, sqTimeSlotMask, nMaxHbusTimeSlot); DemoCheckRc(g_hPipes[iBrd].hHsSpkrPipe[nHs], result);
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[nHsTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsSpkrPipe[nHs], OTX_PIPE_OUTPUT,   g_hPhysDevs[iBrd].hTss, g_demoEnv.nHighwayCodec, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[iBrd].hHsSpkrPipe[nHs], result);
  }
  
  g_sHs[nHs].bConnected = OTX_TRUE;

  return OTX_TRUE;
}



/*******************************************************************************

  @func

*******************************************************************************/
OTX_BOOL DemoDisconnectHsViaH100(
  OTX_INT32 nHs,
  OTX_BOOL bVerbose
)
{
  OTX_INT32 iBrd;
  OTX_INT32 nH100Hwy;
  OTX_INT32 nTs0To1;
  OTX_INT32 nTs1To0;
  OTX_RESULT result;
  
    for (iBrd = 0; iBrd < g_demoEnv.coBrds; iBrd++) {
    if (g_sHs[nHs].bConnected) {
      nH100Hwy = g_sHs[nHs].nH100Hwy;
      nTs0To1  = g_sHs[nHs].nTs0To1;
      nTs1To0  = g_sHs[nHs].nTs1To0;

      if (bVerbose) DemoLogf("Board#%d: Disconnecting HS%d Pipes (nTs0To1=%d, nTs1To0=%d)\n", iBrd, nHs, nTs0To1, nTs1To0);

      result = OtxDrvDisconnectPipe(g_hPipes[iBrd].hHsMicPipe[nHs], OTX_PIPE_INPUT); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[nHs], result);
      result = OtxDrvDisconnectPipe(g_hPipes[iBrd].hHsMicPipe[nHs], OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[nHs], result);
      result = OtxDrvDisconnectPipe(g_hPipes[iBrd].hHsSpkrPipe[nHs], OTX_PIPE_INPUT); DemoCheckRc(g_hPipes[iBrd].hHsSpkrPipe[nHs], result);
      result = OtxDrvDisconnectPipe(g_hPipes[iBrd].hHsSpkrPipe[nHs], OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes[iBrd].hHsSpkrPipe[nHs], result);
      
      g_sH100Hwy[nH100Hwy].bOccupied[nTs0To1] = OTX_FALSE;
      g_sH100Hwy[nH100Hwy].bOccupied[nTs1To0] = OTX_FALSE;
      
    } else {
      DemoLogf("Error: Hs#%d is not connected on Brd#%d\n", nHs, iBrd);
      return OTX_FALSE;
    }
    
  }
  
  return OTX_TRUE;
}

static const char *ms_sqPhysicalDeviceTypeStrings[] = {
  "UNKNOWN",
  "DRIVER",
  "ARNI_PCI",
  "BALDER_PCMCIA",
  "BALDER_S_PCI",
  "BALDER_U_PCI",
  "THOR_PCI",
  "THOR_PCMCIA",
  "VIDAR_5x_PCI",
  "VIDAR_6x_PCI",
  "SIF_PCI",
  "VIDAR_5x_ASM",
  "VIDAR_6x_ASM",
  "VIDAR_55x4_ASM",
  "LI_CONTROLLER",
  "LI_POTS",
  "LI_SLIC",
  "LI_T1E1",
  "LI_BRI_S",
  "LI_BRI_U",
  "HDLC",
  "TSS",
  "DSP",
  "PROCESSOR",
  "PCI_BRIDGE",
  "CODEC",
  "EEPROM",
  "TBC",
  "QUAD_T1E1",
  "QUAD_CODEC",
  "THOR_8_PCI",
  "CT_BUS",
  "HERMOD_PCI",
  "GIMLE_16_PCI",
  "VIDAR_5X1_ASM_BASIC",
  "THOR_1_PCI-BASIC",
  "VIDAR_5X4_ASM_CST",      
  "VIDAR_5X4_ASM_V90",      
  "VIDAR_5X4_ASM_EX",      
  "THOR_8_PCI_PLUS",    
  "THOR_2_PCI_PLUS",    
  "SLEIPNIR_1_PCI",     
  "VIDAR-55X12_PCI",
  "BURST",
  "VIDAR_5X4_ASM_PRO"
};


static const char *ms_sqLogicalDeviceTypeStrings[] = {
  "UNKNOWN",
  "EVENT_QUEUE",
  "PIPE",
  "TONE_CONST_VAL_GENERATOR",
  "TONE_EFFECTS_GENERATOR",
  "TONE_SINEWAVE_GENERATOR",
  "TONE_DTMF_GENERATOR",
  "TONE_MF_GENERATOR",
  "TONE_SEQUENCE_GENERATOR",
  "TONE_DTMF_DIALER",
  "TONE_MF_DIALER",
  "TONE_SILENCE_DETECTOR",
  "TONE_DIALTONE_DETECTOR",
  "TONE_DTMF_DETECTOR",
  "TONE_MF_DETECTOR",
  "TONE_FSK_DETECTOR",
  "DATA_RAW_SENDER",
  "DATA_RAW_RECEIVER",
  "DATA_CONVERTER",
  "DATA_MIXER",
  "DATA_CONFERENCE_BRIDGE",
  "HDLC_SENDER",
  "HDLC_RECEIVER",
  "T1_ROB_BIT_ACCESS",
  "E1_S_BIT_ACCESS",
  "E1_ABCD_BIT_ACCESS",
  "MODEM_V34_SENDER",
  "MODEM_V34_RECEIVER",
  "FAX_V17_SENDER",
  "FAX_V17_RECEIVER",
  "VOICE_ECHO_CANCELLER",
  "VOICE_SILENCE_SUPRESSOR",
  "VOICE_CODEC_G723_ENCODER",
  "VOICE_CODEC_G723_DECODER",
  "VOICE_CODEC_G729_ENCODER",
  "VOICE_CODEC_G729_DECODER",
  "USER_APPLICATION",
  "TONE_VAL_SAMPLER",
  "T1_DL_BIT_ACCESS",
  "TONE_DUAL_DETECTOR",
  "TONE_SINGLE_DETECTOR",
};

     
// Returns the string representing the device type for a specific device, automatically chooses between logical and physical devices
static const char *DemoGetDeviceTypeString(IN OTX_HANDLE hDevice)
{
  OTX_INT32 eDevType;
  assert(sizeof(ms_sqPhysicalDeviceTypeStrings) / sizeof(ms_sqPhysicalDeviceTypeStrings[0]) == (((size_t) OTX_DEVICE_END)));
  assert(sizeof(ms_sqLogicalDeviceTypeStrings) / sizeof(ms_sqLogicalDeviceTypeStrings[0]) == (size_t) OTX_LDEVICE_END);

  if (OTX_SUCCESS(OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, (OTX_UINT32 *)&eDevType))) {
    if (eDevType >= 0) {
      if (eDevType < (OTX_INT32) sizeof(ms_sqPhysicalDeviceTypeStrings)) {
        return ms_sqPhysicalDeviceTypeStrings[eDevType];
      }
    } else if (eDevType < 0) {
      eDevType = -eDevType;
      if (eDevType < (OTX_INT32) sizeof(ms_sqLogicalDeviceTypeStrings)) {
        return ms_sqLogicalDeviceTypeStrings[eDevType];
      }
    }
  }
  return "*** Error ***";
}


// Returns the device no for a specific device
static signed int DemoGetDeviceNo(IN OTX_HANDLE hDevice)
{
  OTX_UINT32 nDevNo;
  if (OTX_SUCCESS(OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDevNo))) {
    return nDevNo;
  } else {
    return -1;
  }
}

// Returns the device no of the host for a specific device
static signed int DemoGetNicDeviceNo(IN OTX_HANDLE hDevice)
{
  OTX_HANDLE hNicDevice;
  OTX_UINT32 nDevNo;

  if (OTX_SUCCESS(OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NIC, &hNicDevice))) {
    if (OTX_SUCCESS(OtxDrvGetAttributeValueUint32(hNicDevice, OTX_ATTR_DEVICE_NO, &nDevNo))) {
      return nDevNo;
    } else {
      return -1;
    }
  } else {
    return -1;
  }

}




/*******************************************************************************

  @func

*******************************************************************************/
void DemoPrintPipeConnections(
  OTX_HANDLE hPipe
)
{
  OTX_RESULT result;
  OTX_UINT32 nCapacity;
  OTX_UINT32 nInHandle;
  OTX_UINT32 nOutHandle;
  OTX_UINT32 nInHighway;
  OTX_UINT32 nOutHighway;
  OTX_UINT32 cbReturned;
  OTX_UINT8 sqInTimeSlotMask[128];
  OTX_UINT8 sqOutTimeSlotMask[128];
  OTX_UINT32 n;
    
  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_CAPACITY, &nCapacity); DemoCheckRc(hPipe, result);
  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_IN_CONNECTED, &nInHandle); DemoCheckRc(hPipe, result);
  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_OUT_CONNECTED, &nOutHandle); DemoCheckRc(hPipe, result);
  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_IN_HIGHWAY, &nInHighway); DemoCheckRc(hPipe, result);
  result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_OUT_HIGHWAY, &nOutHighway); DemoCheckRc(hPipe, result);
  result = OtxDrvGetAttributeValue(hPipe, OTX_ATTR_PIPE_IN_CONN_MASK, sqInTimeSlotMask, 128, &cbReturned); DemoCheckRc(hPipe, result);
  result = OtxDrvGetAttributeValue(hPipe, OTX_ATTR_PIPE_OUT_CONN_MASK, sqOutTimeSlotMask, 128, &cbReturned); DemoCheckRc(hPipe, result);

  DemoLogf("Connection info for pipe %X (Capacity:%d)\n", hPipe, nCapacity);

  DemoLogf("  Local input  connections: Brd:%d %s[%d] Hwy:%d ", DemoGetNicDeviceNo((OTX_HANDLE)nInHandle), DemoGetDeviceTypeString((OTX_HANDLE)nInHandle), DemoGetDeviceNo((OTX_HANDLE)nInHandle), nInHighway);

  DemoLogf("TS: ");
  for (n = 0; n < 32; n++) {
    if (sqInTimeSlotMask[n]) {
      DemoLogf("%d ", n);
    }
  }
  DemoLogf("\n");

  DemoLogf("  Local output connections: Brd:%d %s[%d] Hwy:%d ", DemoGetNicDeviceNo((OTX_HANDLE)nOutHandle), DemoGetDeviceTypeString((OTX_HANDLE)nOutHandle), DemoGetDeviceNo((OTX_HANDLE)nOutHandle), nOutHighway);

  DemoLogf("TS: ");
  for (n = 0; n < 32; n++) {
    if (sqOutTimeSlotMask[n]) {
      DemoLogf("%d ", n);
    }
  }
  DemoLogf("\n");

  // CT Bus connection info
  {
    OtxPipeCtBusConnectionS *psqCtBusConnectPoints;
    OTX_UINT32  cbCtBusConnectPoints;  
    OTX_UINT32  coCtBusConnectPoints;  
    OTX_UINT32  cbAttribValue;   // Bytes actually used in psqCtBusConnectPoints

    result = OtxDrvGetAttributeValueUint32(hPipe, OTX_ATTR_PIPE_CT_BUS_CONNECTIONS_COUNT, &coCtBusConnectPoints); DemoCheckRc(hPipe, result);
    cbCtBusConnectPoints = (coCtBusConnectPoints * sizeof(OtxPipeCtBusConnectionS));

    psqCtBusConnectPoints = (OtxPipeCtBusConnectionS *) malloc(cbCtBusConnectPoints * sizeof(char));
    assert(psqCtBusConnectPoints);

    result = OtxDrvGetAttributeValue(hPipe, OTX_ATTR_PIPE_CT_BUS_CONNECTIONS_DETAILS, psqCtBusConnectPoints, cbCtBusConnectPoints, &cbAttribValue); DemoCheckRc(hPipe, result);
    assert(cbAttribValue == cbCtBusConnectPoints);

    if (coCtBusConnectPoints) {
      for (n = 0; n < coCtBusConnectPoints; n++) {
        DemoLogf("  Sourcing CT Bus:%d, CT Highway:%d, CT TS:%d\n", psqCtBusConnectPoints[n].nCtBus, psqCtBusConnectPoints[n].nHighway, psqCtBusConnectPoints[n].nTimeslot );
      }
    } else {
      DemoLogf("  No sourced CT Bus connections\n");
    }

    free(psqCtBusConnectPoints);
  }

    
}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->law               = OTX_DATA_FORMAT_COMPAND8_A_LAW;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->bSs7FisuFiltering = OTX_FALSE;
  envP->coCtBuses         = 1;
  envP->coBrds            = 2;
  envP->nHbusRate         = 2048;
  envP->eClkSource        = OTX_CLOCK_SOURCE_INTERNAL;
  envP->eClkMode          = OTX_CLOCK_MODE_H100;
  envP->nBoardType        = OTX_DEVICE_THOR_PCI;


  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          envP->logFileName = argv[i]+2;
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'C') {   // Companding Law
          if (*(argv[i]+2) == 'A') {   // A-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') {   // U-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_THOR_PCI;
            envP->nHighwayCodec = 14;
          } else if ( *(argv[i]+2) == '8') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
            envP->nHighwayCodec = 14;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->nHighwayCodec = 0;    // codec is highway 0 on the 8Plus
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }

        } else if (*(argv[i]+1) == 'K') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   // Line Interface 0
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'J') {   // Clock Mode
          if (*(argv[i]+2) == 'H') {   
            envP->eClkMode = OTX_CLOCK_MODE_H100;
          } else if (*(argv[i]+2) == 'E') {  
            envP->eClkMode = OTX_CLOCK_MODE_H100_ETCF;
          } else {
            DemoLogf("Error: Invalid Clock Mode command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'N') {   // Number of CT buses
          envP->coCtBuses = strtol(argv[i]+2, NULL, 0);
          envP->coBrds = 2*envP->coCtBuses;  // There should be two boards connected per CT Bus in this example

        } else if (*(argv[i]+1) == 'H') {   // Hbus Rate
          envP->nHbusRate = strtol(argv[i]+2, NULL, 0);
          switch (envP->nHbusRate) {
          case 2048:
          case 4096:
          case 8192:
            break; // Acceptable value
          
          default:
            DemoLogf("Error: Invalid Hbus Rate %d. Valid rates are 2048, 4096 and 8192 kbps.\n", envP->nHbusRate);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }

        } else if (*(argv[i]+1) == 'S') {   // SS7 FISU filtering
          envP->bSs7FisuFiltering = OTX_TRUE;

        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /N<CtBuses>  : Number of CT Buses (default is 1)\n");
  DemoLogf(" /T[2|8|P]    : Board Type used for the demo, where\n");
  DemoLogf("                  2 = Use Thor-2-PCI Board\n");
  DemoLogf("                  8 = Use Thor-8-PCI Board\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf(" /H<Rate>     : H.100 Bus Rate. Valid rates are 2048, 4096 and 8192 (default is 2048) kbps\n");
  DemoLogf(" /S           : SS7 FISU filtering active\n");
  DemoLogf(" /CA          : A-law\n");
  DemoLogf(" /CU          : u-law\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /KA          : Use Internal board oscillator as clock source\n");
  DemoLogf(" /K0          : Use recovered clock from Line Interface 0 as clock source\n");
  DemoLogf(" /JH          : Use H100 MC-1 as the clock mode\n");
  DemoLogf(" /JE          : Use H100 ETCF as the clock mode\n");
  DemoLogf(" /Z<str>      : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  OTX_RESULT result;
  static OTX_INT32      nCurrentBrd = 0;
  static OTX_INT32      nCurrentLi = 0;
  static OTX_INT32      nCurrentHs = 2;
  static OTX_INT32      nCurrentTs0To1 = 5;
  OTX_INT32 nMaxLocalHw;

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
    nMaxLocalHw = 16;
  } else {
    nMaxLocalHw = 8;
  }

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" x   Connect HS2 (Brd0 and Brd1) to H.100 Highway 1, using device<->TSS connection\n");
    DemoLogf(" X   Disconnect HS2 (Brd0 and Brd1) (device<->TSS connection).\n");
    DemoLogf(" p   Connect HS2:Brd0 to HS2:Brd1 (using CT Bus connections)\n");
    DemoLogf(" P   Disconnect HS2:Brd0 to HS2:Brd1 (using CT Bus connections)\n");
    DemoLogf(" l   Query Line Status of both Lis on all boards\n");
    DemoLogf(" 0-1 Select active Li, this Li is then used for subsequent commands\n");
    DemoLogf(" b   Increment Current Board (current is Brd%d) \n", nCurrentBrd);
    DemoLogf(" B   Decrement Current Board (current is Brd%d) \n", nCurrentBrd);
    DemoLogf(" m   Increment Current Handset (current is Hs%d) \n", nCurrentHs);
    DemoLogf(" M   Decrement Current Handset (current is Hs%d) \n", nCurrentHs);
    DemoLogf(" n   Increment starting TS for next Handset connection (nCurrentTs0To1 is %d) \n", nCurrentTs0To1);
    DemoLogf(" N   Decrement starting TS for next Handset connection (nCurrentTs0To1 is %d) \n", nCurrentTs0To1);
    DemoLogf(" e   Make 1000 connections/disconnections using Handset 1,2 and 3. Random TS.\n");
    DemoLogf(" z   Connecting/disconnecting maximum number of pipes (%d pipes) to the H.100 bus\n", nMaxLocalHw * 32);
    DemoLogf(" k   Set H-Bus Rate. Should only be executed when there are no H.100 connections\n");
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;
      OTX_INT32 iBrd;

      for (iBrd = 0; iBrd < g_demoEnv.coBrds; iBrd++) {
        for (nLi = 0; nLi < 2; nLi++) {
          DemoLogf("Brd#%d: Line status of Li #%d is ", iBrd, nLi);
          result = OtxT1E1GetLayer1Status(g_hPhysDevs[iBrd].hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs[iBrd].hLi[nLi], result);
          DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
        }
      }
    }
    break;

  case 'a':
    {
      OTX_INT32 nHs = nCurrentHs;
      OTX_INT32 nH100Hwy = 1;
      OTX_INT32 nTs0To1 = nCurrentTs0To1;
      OTX_INT32 nTs1To0 = (nCurrentTs0To1 + 1) % 32;
      DemoLogf("Connecting Hs%d to nTs0To1=%d, nTs1To0=%d \n", nHs, nTs0To1, nTs1To0);

      DemoConnectHsViaH100(nHs, nH100Hwy, nTs0To1, nTs1To0, OTX_TRUE);

      DemoLogf("Done.\n");
    }
    break;


  case 'A':
    {
      OTX_INT32 nHs = nCurrentHs;
      DemoLogf("Disconnecting Hs%d \n", nHs);

      DemoDisconnectHsViaH100(nHs, OTX_TRUE);

      DemoLogf("Done.\n");
    }
    break;

  case 'p':
    {
      OTX_INT32 nHs = nCurrentHs;
      OTX_INT32 nCtBus = nCurrentBrd/2;
      OTX_INT32 nBrdA = nCtBus*2+0;
      OTX_INT32 nBrdB = nCtBus*2+1;
      
      // Connect/Disconnect using the board-to-board CT Bus pipe connection scheme
      DemoLogf("Connecting HS%d on board#%d to HS%d on board#%d over CtBus#%d\n", nHs, nBrdA, nHs, nBrdB, nCtBus);
      result = OtxDrvConnectPipe(g_hPipes[nBrdA].hHsMicPipe[nHs], OTX_PIPE_INPUT,   g_hPhysDevs[nBrdA].hCodec[nHs], -1, NULL, 0); DemoCheckRc(g_hPipes[nBrdA].hHsMicPipe[nHs], result);
      result = OtxDrvConnectPipe(g_hPipes[nBrdA].hHsMicPipe[nHs], OTX_PIPE_OUTPUT,  g_hPhysDevs[nBrdB].hCodec[nHs], -1, NULL, 0); DemoCheckRc(g_hPipes[nBrdA].hHsMicPipe[nHs], result);

      DemoLogf("Pipe for Hs%d[%d] -> Hs%d[%d]: \n", nHs, nBrdA, nHs, nBrdB); DemoPrintPipeConnections(g_hPipes[nBrdA].hHsMicPipe[nHs]);
      
      result = OtxDrvConnectPipe(g_hPipes[nBrdA].hHsSpkrPipe[nHs], OTX_PIPE_INPUT,   g_hPhysDevs[nBrdB].hCodec[nHs], -1, NULL, 0); DemoCheckRc(g_hPipes[nBrdA].hHsSpkrPipe[nHs], result);
      result = OtxDrvConnectPipe(g_hPipes[nBrdA].hHsSpkrPipe[nHs], OTX_PIPE_OUTPUT,  g_hPhysDevs[nBrdA].hCodec[nHs], -1, NULL, 0); DemoCheckRc(g_hPipes[nBrdA].hHsSpkrPipe[nHs], result);
      
      DemoLogf("Pipe for Hs%d[%d] -> Hs%d[%d]: \n", nHs, nBrdB, nHs, nBrdA); DemoPrintPipeConnections(g_hPipes[nBrdA].hHsSpkrPipe[nHs]);

      DemoLogf("Done.\n");
    }
    break;
  case 'P':
    {
      OTX_INT32 nHs = nCurrentHs;
      OTX_INT32 nCtBus = nCurrentBrd/2;
      OTX_INT32 nBrdA = nCtBus*2+0;
      OTX_INT32 nBrdB = nCtBus*2+1;
      
      DemoLogf("Disconnecting HS%d on board#%d from HS%d on board#%d over CtBus#%d\n", nHs, nBrdA, nHs, nBrdB, nCtBus);
      result = OtxDrvDisconnectPipe(g_hPipes[nBrdA].hHsMicPipe[nHs], OTX_PIPE_INPUT); DemoCheckRc(g_hPipes[nBrdA].hHsMicPipe[nHs], result);
      result = OtxDrvDisconnectPipe(g_hPipes[nBrdA].hHsMicPipe[nHs], OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes[nBrdA].hHsMicPipe[nHs], result);

      result = OtxDrvDisconnectPipe(g_hPipes[nBrdA].hHsSpkrPipe[nHs], OTX_PIPE_INPUT); DemoCheckRc(g_hPipes[nBrdA].hHsSpkrPipe[nHs], result);
      result = OtxDrvDisconnectPipe(g_hPipes[nBrdA].hHsSpkrPipe[nHs], OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes[nBrdA].hHsSpkrPipe[nHs], result);
      
      DemoLogf("Done.\n");
    }
    break;

  case 'x':
    {
      OTX_INT32 nHs = nCurrentHs;

      OTX_UINT8 sqTimeSlotMask[128];
      OTX_INT32 ts;
      OTX_INT32 iBrd;
      OTX_INT32 nMaxHbusTimeSlot = g_demoEnv.nHbusRate/64;

      for (iBrd = 0; iBrd < 2; iBrd++) {
        
        DemoLogf("Board#%d: Connecting HS2 to H100 Hwy1, ", iBrd);
        ts = 3; // HS0 is TS1, HS1 is TS2, HS2 is TS3, HS3 is TS4 on the g_demoEnv.nHighwayCodec
        memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
        sqTimeSlotMask[ts] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsMicPipe[nHs], OTX_PIPE_INPUT,   g_hPhysDevs[iBrd].hTss, g_demoEnv.nHighwayCodec, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[nHs], result);

        if (iBrd == 0) {
          ts = nMaxHbusTimeSlot-2;
        } else {
          ts = nMaxHbusTimeSlot-1;
        }
        DemoLogf("Mic to TS%d, ", ts);
        memset(sqTimeSlotMask, 0, nMaxHbusTimeSlot * sizeof(OTX_UINT8));
        sqTimeSlotMask[ts] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsMicPipe[nHs], OTX_PIPE_OUTPUT,   g_hPhysDevs[iBrd].hTss, DEMO_H100_HWY_1, sqTimeSlotMask, nMaxHbusTimeSlot); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[nHs], result);
        
        if (iBrd == 0) {
          ts = nMaxHbusTimeSlot-1;
        } else {
          ts = nMaxHbusTimeSlot-2;
        }
        DemoLogf("Spkr to TS%d \n", ts);
        memset(sqTimeSlotMask, 0, nMaxHbusTimeSlot * sizeof(OTX_UINT8));
        sqTimeSlotMask[ts] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsSpkrPipe[nHs], OTX_PIPE_INPUT,   g_hPhysDevs[iBrd].hTss, DEMO_H100_HWY_1, sqTimeSlotMask, nMaxHbusTimeSlot); DemoCheckRc(g_hPipes[iBrd].hHsSpkrPipe[nHs], result);
        
        ts = 3;
        memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
        sqTimeSlotMask[ts] = 0xFF;
        result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsSpkrPipe[nHs], OTX_PIPE_OUTPUT,   g_hPhysDevs[iBrd].hTss, g_demoEnv.nHighwayCodec, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[iBrd].hHsSpkrPipe[nHs], result);
      }

      DemoLogf("Done.\n");
    }
    break;

  case 'X':
    {
      OTX_INT32 iBrd;
      OTX_INT32 nHs = nCurrentHs;

      for (iBrd = 0; iBrd < 2; iBrd++) {
        DemoLogf("Board#%d: Disconnecting HS2 Pipes (Spkr and Mic)\n", iBrd);
        result = OtxDrvDisconnectPipe(g_hPipes[iBrd].hHsMicPipe[nHs], OTX_PIPE_INPUT); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[nHs], result);
        result = OtxDrvDisconnectPipe(g_hPipes[iBrd].hHsMicPipe[nHs], OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[nHs], result);
        result = OtxDrvDisconnectPipe(g_hPipes[iBrd].hHsSpkrPipe[nHs], OTX_PIPE_INPUT); DemoCheckRc(g_hPipes[iBrd].hHsSpkrPipe[nHs], result);
        result = OtxDrvDisconnectPipe(g_hPipes[iBrd].hHsSpkrPipe[nHs], OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes[iBrd].hHsSpkrPipe[nHs], result);

      }
      
      DemoLogf("Done.\n");
    }
    break;


  case 'k':
    {
      char szBuf[12];
      OTX_INT32 iBrd;
      OTX_UINT32 nHbusRate;
      OTX_ATTRIB_ID nAttribId;
      
      DemoLogf("Set H-Bus Rate. Which H.100 Stream Group?\n");
      
      DemoLogf("  a: Stream Group A (d0-d3)\n");
      DemoLogf("  b: Stream Group B (d4-d7)\n");
      DemoLogf("  c: Stream Group C (d8-d11)\n");
      DemoLogf("  d: Stream Group D (d12-d15)\n");
      DemoLogf("  e: Stream Group E (d16-d19)\n");
      DemoLogf("  f: Stream Group F (d20-d23)\n");
      DemoLogf("  g: Stream Group G (d24-d27)\n");
      DemoLogf("  h: Stream Group H (d28-d31)\n");
      
      OsaConsoleGetStr(szBuf);
      
      nAttribId = OTX_ATTR_TSS_CTBUS_RATE_STREAM_GROUP_A + (*szBuf - 'a');
      
      for (iBrd = 0; iBrd < 2; iBrd++) {
        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[iBrd].hTss, nAttribId, &nHbusRate); DemoCheckRc(g_hPhysDevs[iBrd].hTss, result);
        
        DemoLogf("Brd#%d: Stream Group %s is currently configured for rate = %d kbps\n", iBrd, szBuf, nHbusRate);
      }
      
      DemoLogf("Select new H-Bus Rate? (2048, 4096, or 8192) ");
      
      OsaConsoleGetStr(szBuf);
      if (*szBuf) {
        nHbusRate = (OTX_UINT32) strtoul(szBuf, NULL, 0);
        for (iBrd = 0; iBrd < 2; iBrd++) {
          DemoLogf("Brd#%d: Setting rate to %d.\n", iBrd, nHbusRate);
          result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[iBrd].hTss, nAttribId, nHbusRate); DemoCheckRc(g_hPhysDevs[iBrd].hTss, result);
        }
      } else {
        DemoLogf("Rate remains unchanged.\n");
      }
      
      DemoLogf("Done.\n");
    }
    break;


  case 'z':
    {
      OTX_INT32 nMaxHbusTimeSlot = g_demoEnv.nHbusRate/64;
      OTX_HANDLE hsqFromT1E1Pipes[16][32];
      OTX_INT32 iBrd = 0;
      OTX_UINT8 sqLocalTimeSlotMask[32];
      OTX_UINT8 sqHbusTimeSlotMask[128];

      DemoLogf("Connecting maximum number of pipes to the H.100 bus\n");

      // Connect pipes from the E1T1 device to the Hbus
      {
        OTX_INT32 nLocalHw;

        for (nLocalHw = 0; nLocalHw < nMaxLocalHw; nLocalHw++) {
          OTX_INT32 nLocalTs;
          
          for (nLocalTs = 0; nLocalTs < 32; nLocalTs++) {
            OTX_INT32 nHbusHw = DEMO_H100_HWY_0 + (nLocalHw*32+nLocalTs)/nMaxHbusTimeSlot;
            OTX_INT32 nHbusTs = (nLocalHw*32+nLocalTs) % nMaxHbusTimeSlot;
            
            result = OtxDrvCreatePipe(64, &hsqFromT1E1Pipes[nLocalHw][nLocalTs]); DemoCheckRc(hsqFromT1E1Pipes[nLocalHw][nLocalTs], result);
            
            DemoLogf("Connecting Local Highway %d, TS%d -> ", nLocalHw, nLocalTs);

            memset(sqLocalTimeSlotMask, 0, sizeof(sqLocalTimeSlotMask));
            sqLocalTimeSlotMask[nLocalTs] = 0xFF;
            
            result = OtxDrvConnectPipe(hsqFromT1E1Pipes[nLocalHw][nLocalTs], OTX_PIPE_INPUT, g_hPhysDevs[iBrd].hTss, nLocalHw, sqLocalTimeSlotMask, 32); DemoCheckRc(hsqFromT1E1Pipes[nLocalHw][nLocalTs], result);

            DemoLogf("Hbus Highway %d, TS%d\n", nHbusHw, nHbusTs);

            memset(sqHbusTimeSlotMask, 0, sizeof(sqHbusTimeSlotMask));
            sqHbusTimeSlotMask[nHbusTs] = 0xFF;

            result = OtxDrvConnectPipe(hsqFromT1E1Pipes[nLocalHw][nLocalTs], OTX_PIPE_OUTPUT, g_hPhysDevs[iBrd].hTss, nHbusHw, sqHbusTimeSlotMask,128); DemoCheckRc(hsqFromT1E1Pipes[nLocalHw][nLocalTs], result);
          }
          
        }
      }


      DemoLogf("All pipes connected. Then press any key to close (disconnect) them...\n");
      OsaConsoleGetCh();

      {
        OTX_INT32 nLocalHw;

        for (nLocalHw = 0; nLocalHw < nMaxLocalHw; nLocalHw++) {
          OTX_INT32 nLocalTs;
          
          for (nLocalTs = 0; nLocalTs < 32; nLocalTs++) {
            result = OtxDrvClose(hsqFromT1E1Pipes[nLocalHw][nLocalTs]); DemoCheckRc(hsqFromT1E1Pipes[nLocalHw][nLocalTs], result);
          }
        }
      }

      DemoLogf("Done.\n");
    }
    break;


#if 0 // for driver debugging
  case 'j':
    {
      OTX_INT32 nHs = nCurrentHs;
      
      OTX_UINT8 sqLocalTimeSlotMask[32];
      OTX_UINT8 sqHbusTimeSlotMask[128];
      OTX_INT32 ts;
      OTX_INT32 iBrd;
      
      
      iBrd = 0;
      ts = 3; // HS0 is TS1, HS1 is TS2, HS2 is TS3, HS3 is TS4 on the g_demoEnv.nHighwayCodec
      DemoLogf("Board#%d: Connecting HS2 to H100 Hwy1, ", iBrd);
      memset(sqLocalTimeSlotMask, 0, sizeof(sqLocalTimeSlotMask));
      sqLocalTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsMicPipe[nHs], OTX_PIPE_INPUT,   g_hPhysDevs[iBrd].hTss, g_demoEnv.nHighwayCodec, sqLocalTimeSlotMask, 32); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[nHs], result);
      
      iBrd = 0;
      ts = 127;
      DemoLogf("Mic to TS%d, ", ts);
      memset(sqHbusTimeSlotMask, 0, sizeof(sqHbusTimeSlotMask));
      sqHbusTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsMicPipe[nHs], OTX_PIPE_OUTPUT,   g_hPhysDevs[iBrd].hTss, DEMO_H100_HWY_1, sqHbusTimeSlotMask, 128); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[nHs], result);
      
      iBrd = 1;
      ts = 127;
      DemoLogf("Spkr to TS%d \n", ts);
      memset(sqHbusTimeSlotMask, 0,  sizeof(sqHbusTimeSlotMask));
      sqHbusTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsSpkrPipe[nHs], OTX_PIPE_INPUT,   g_hPhysDevs[iBrd].hTss, DEMO_H100_HWY_1, sqHbusTimeSlotMask, 128); DemoCheckRc(g_hPipes[iBrd].hHsSpkrPipe[nHs], result);
      
      iBrd = 1;
      ts = 3;
      memset(sqLocalTimeSlotMask, 0, sizeof(sqLocalTimeSlotMask));
      sqLocalTimeSlotMask[ts] = 0xFF;
      result = OtxDrvConnectPipe(g_hPipes[iBrd].hHsSpkrPipe[nHs], OTX_PIPE_OUTPUT,   g_hPhysDevs[iBrd].hTss, g_demoEnv.nHighwayCodec, sqLocalTimeSlotMask, 32); DemoCheckRc(g_hPipes[iBrd].hHsSpkrPipe[nHs], result);
      
      DemoLogf("Done.\n");
    }
    break;
  case 'J':
    {
      OTX_INT32 iBrd;
      OTX_INT32 nHs = nCurrentHs;
      
      iBrd = 0;
      DemoLogf("Board#%d: Disconnecting HS2 Pipes (Spkr and Mic)\n", iBrd);
      result = OtxDrvDisconnectPipe(g_hPipes[iBrd].hHsMicPipe[nHs], OTX_PIPE_INPUT); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[nHs], result);
      result = OtxDrvDisconnectPipe(g_hPipes[iBrd].hHsMicPipe[nHs], OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes[iBrd].hHsMicPipe[nHs], result);
      
      iBrd = 1;
      DemoLogf("Board#%d: Disconnecting HS2 Pipes (Spkr and Mic)\n", iBrd);
      result = OtxDrvDisconnectPipe(g_hPipes[iBrd].hHsSpkrPipe[nHs], OTX_PIPE_INPUT); DemoCheckRc(g_hPipes[iBrd].hHsSpkrPipe[nHs], result);
      result = OtxDrvDisconnectPipe(g_hPipes[iBrd].hHsSpkrPipe[nHs], OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes[iBrd].hHsSpkrPipe[nHs], result);
      
      DemoLogf("Done.\n");
    }
    break;
#endif    
    


  case 'b':
    {
      // Select current nCurrentBrd

      nCurrentBrd++;
      if (nCurrentBrd > (g_demoEnv.coBrds-1)) {
        // Wrap 
        nCurrentBrd = 0;
      }

      DemoLogf("Brd%d Selected\n", nCurrentBrd);
    }
    break;

  case 'B':
    {
      // Select current nCurrentBrd
      nCurrentBrd--;
      if (nCurrentBrd < 0) {
        // Wrap 
        nCurrentBrd = (g_demoEnv.coBrds-1);
      }
      DemoLogf("Brd%d Selected\n", nCurrentBrd);
    }
    break;

  case 'm':
    {
      // Select current nCurrentHs

      nCurrentHs++;
      if (nCurrentHs > 3) {
        // Wrap 
        nCurrentHs = 0;
      }

      DemoLogf("Handset%d Selected\n", nCurrentHs);
    }
    break;

  case 'M':
    {
      // Select current nCurrentHs
      nCurrentHs--;
      if (nCurrentHs < 1) {
        // Wrap 
        nCurrentHs = 3;
      }
      DemoLogf("Handset%d Selected\n", nCurrentHs);
    }
    break;

  case 'n':
    {
      // Select current nCurrentTs0To1

      nCurrentTs0To1++;
      if (nCurrentTs0To1 > 31) {
        // Wrap 
        nCurrentTs0To1 = 0;
      }

      DemoLogf("nTs0To1 = %d\n", nCurrentTs0To1);
    }
    break;

  case 'N':
    {
      // Select current nCurrentTs0To1
      nCurrentTs0To1--;
      if (nCurrentTs0To1 < 1) {
        // Wrap 
        nCurrentTs0To1 = 31;
      }
      DemoLogf("nTs0To1 = %d\n", nCurrentTs0To1);
    }
    break;

  case 't':
    {
      OTX_UINT8 sqHdlcFrame[64];
      OTX_INT32 i;
      OTX_INT32 cbHdlcFrame = 10;
      OTX_HANDLE hHandle;

      hHandle = g_hLogicDevs[nCurrentBrd].hHdlcSender[nCurrentLi];

      for (i = 0; i < cbHdlcFrame; i++) {
        sqHdlcFrame[i] = (OTX_UINT8) i;
      }

      DemoLogf("Brd#%d: Sending message on Li #%d ...", nCurrentBrd, nCurrentLi);

      // Send the HDLC frame
      result = OtxHdlcSendData(hHandle, sqHdlcFrame, cbHdlcFrame, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;


  case '0':
  case '1':
    // Select current Li 
    nCurrentLi = commandChar - '0';
    DemoLogf("Li#%d Selected ....\n", nCurrentLi);
    break;

  case 'r':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Brd#%d: Read from an 8-bit register in Li%d\n", nCurrentBrd, nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs[nCurrentBrd].hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hLi[nCurrentLi], result);

      DemoLogf("Brd#%d: Li%d Address 0x%X = %.2X\n", nCurrentBrd, nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'w':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Brd#%d: Write to an 8-bit register in Li%d\n", nCurrentBrd, nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs[nCurrentBrd].hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hLi[nCurrentLi], result);

      DemoLogf("Brd#%d: Wrote %.2X to Li%d Address 0x%X\n", nCurrentBrd, (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

#if 1 // Only for driver debugging
  case 'o':
    {
      char szBuf[12];
      OTX_UINT32 nCmd;
      OTX_UINT8 nTs;
      OTX_UINT8 nCtrlStream;
      
      DemoLogf("Brd#%d: Amb Driver Debug Command. Which Cmd?\n", nCurrentBrd);
      DemoLogf("0: Dump Local Connection Memory\n");
      DemoLogf("1: Dump all three CAMs\n");
      DemoLogf("2: Clear Connections in CAM-Os\n");
      DemoLogf("3: Find CAM entry in CAM-Os\n");
      OsaConsoleGetStr(szBuf);
      nCmd = (OTX_INT32) strtol(szBuf, NULL, 0);

      if (nCmd == 3) {
        DemoLogf("Enter nTs: ");
        OsaConsoleGetStr(szBuf);
        nTs = (OTX_UINT8) strtol(szBuf, NULL, 0);

        DemoLogf("Enter nCtrlStream: ");
        OsaConsoleGetStr(szBuf);
        nCtrlStream = (OTX_UINT8) strtol(szBuf, NULL, 0);

        nCmd |= (nCtrlStream << 3*8) | (nTs << 2*8);
      }
      
      DemoLogf("Brd#%d: Issuing Driver Debug Command %X\n", nCurrentBrd, nCmd);
      OtxDrvDbgDump(g_hPhysDevs[nCurrentBrd].hTss, nCmd);
      DemoLogf("Brd#%d: Done.\n", nCurrentBrd);
    }
    break;
#endif

#if 1 // for driver debugging
  case 'e':
    {
      OTX_INT32 i;
      OTX_INT32 j;
      OTX_INT32 nHs;
      OTX_INT32 nMaxHbusTimeSlot = g_demoEnv.nHbusRate/64;

      for (i = 0; i < 1000; i++) {
        for (nHs = 1; nHs <= 3; nHs++) {
          for (j = 0; j < 100; j++) {
            OTX_INT32 nH100Hwy = rand() % 32;
            OTX_INT32 nTs0To1 = rand() % nMaxHbusTimeSlot;
            OTX_INT32 nTs1To0 = rand() % nMaxHbusTimeSlot;
            if (DemoConnectHsViaH100(nHs, nH100Hwy, nTs0To1, nTs1To0, OTX_FALSE)) {
              DemoLogf("Connected Hs%d to nTs0To1=%d, nTs1To0=%d on highway=%d (attempt %d)\n", nHs, nTs0To1, nTs1To0, nH100Hwy, j+1);
              break;
            }
          }
          if (j >= 100) {
            DemoLogf("ERROR: Could not connect Hs%d (after %d attempts)\n", nHs, j);
            DemoCheckRc(OTX_INVALID_HANDLE_VALUE, OTX_E_FAIL);
          }
        }

        if (i%100 == 0) {
          DemoLogf("Please verify handset connections. Then press any key...\n");
          OsaConsoleGetCh();
        }

        for (nHs = 1; nHs <= 3; nHs++) {
          if (DemoDisconnectHsViaH100(nHs, OTX_FALSE)) {
            DemoLogf("Disconnected Hs%d \n", nHs);
          } else {
            DemoLogf("ERROR: Could not connect Hs%d (after %d attempts)\n", nHs, j);
            DemoCheckRc(OTX_INVALID_HANDLE_VALUE, OTX_E_FAIL);
          }
        }

      }

      DemoLogf("Done.\n");
    }
    break;
#endif


  case '{':
    {
      char strBuf[12];
      static OTX_UINT8 nTssRegister = 0;
      OTX_UINT32 nReadVal;

      DemoLogf("Brd#%d: Read from TSS control register. Which control register? [0x%.2X] ", nCurrentBrd, nTssRegister);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTssRegister = (OTX_UINT8)strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs[nCurrentBrd].hTss, 0x01, nTssRegister); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hTss, result);
      result = OtxDrvWriteReg8(g_hPhysDevs[nCurrentBrd].hTss, 0x02, 0x00); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hTss, result);

      result = OtxDrvReadReg8(g_hPhysDevs[nCurrentBrd].hTss,  0x03, &nReadVal);  DemoCheckRc(g_hPhysDevs[nCurrentBrd].hTss, result);

      DemoLogf("Brd#%d: Read 0x%.2X from TSS Control Register 0x%.2X\n", nCurrentBrd, (OTX_UINT8)nReadVal, nTssRegister);
      DemoLogf("Done.\n");
    }
    break;

  case '}':
    {
      char strBuf[12];
      static OTX_UINT8 nTssRegister = 0;
      static OTX_UINT8 nWriteVal = 0;

      DemoLogf("Brd#%d: Write to TSS control register. Which control register? [0x%.2X] ", nCurrentBrd, nTssRegister);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTssRegister = (OTX_UINT8)strtoul(strBuf, NULL, 0);
      }

      DemoLogf("Brd#%d: Which value? [0x%.2X] ", nCurrentBrd, nWriteVal);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nWriteVal = (OTX_UINT8)strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs[nCurrentBrd].hTss, 0x01, nTssRegister); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hTss, result);
      result = OtxDrvWriteReg8(g_hPhysDevs[nCurrentBrd].hTss, 0x02, 0x00); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hTss, result);

      result = OtxDrvWriteReg8(g_hPhysDevs[nCurrentBrd].hTss,  0x03, nWriteVal);  DemoCheckRc(g_hPhysDevs[nCurrentBrd].hTss, result);

      DemoLogf("Brd#%d: Wrote 0x%.2X to TSS Control Register 0x%.2X\n", nCurrentBrd, nWriteVal, nTssRegister);
      DemoLogf("Done.\n");
    }
    break;


  case 'c':
    {
      OTX_UINT8 highWay;
      OTX_UINT8 channel;
      OTX_UINT8 byteVal;
      char wValStr[12];
      
      DemoLogf("Brd#%d: Set Constant Byte on Channel\n", nCurrentBrd);
      
      DemoLogf("Highway: ");
      OsaConsoleGetStr(wValStr);
      highWay = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
      DemoLogf("Channel: ");
      OsaConsoleGetStr(wValStr);
      channel = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
      DemoLogf("Constant Byte: ");
      OsaConsoleGetStr(wValStr);
      byteVal = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
      result = OtxBrdTssConstByte(g_hPhysDevs[nCurrentBrd].hTss, highWay, channel, byteVal); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hTss, result);

      DemoLogf("Done.\n");
    }
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;
  OTX_INT32 iBrd;

	OsaConsoleInit();

  DemoLogf("2T2P, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Thor-2-PCI Example program.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
		OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is        : %s\n", g_demoEnv.logFileName);
  DemoLogf("Number of CT Buses : %d\n", g_demoEnv.coCtBuses);
  DemoLogf("Number of boards   : %d\n", g_demoEnv.coBrds);

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();

  // Connect to the Physical CT Buses
  {
    OTX_INT32 nCtBus;
    for (nCtBus = 0; nCtBus < g_demoEnv.coCtBuses; nCtBus++) {
      DemoLogf("Open OTX_DEVICE_CT_BUS[%d]\n", nCtBus);
      result = OtxDrvOpenPhysicalDevice(0, OTX_DEVICE_CT_BUS, nCtBus, nCtBus, g_hEventQueue, &g_hCtBus[nCtBus]); DemoCheckRc(g_hCtBus[nCtBus], result);

      // Limit the number of highways that OTX can use on this CT Bus (only to demonstrate this functionality)
      result = OtxDrvSetAttributeValueUint32(g_hCtBus[nCtBus], OTX_ATTR_CTBUS_OTX_HWYS, 4); DemoCheckRc(g_hCtBus[nCtBus], result);

      // Limit the number of timeslots that OTX can use on this CT Bus (only to demonstrate this functionality)
      result = OtxDrvSetAttributeValueUint32(g_hCtBus[nCtBus], OTX_ATTR_CTBUS_OTX_TIMESLOTS_PER_HWY, 16); DemoCheckRc(g_hCtBus[nCtBus], result);

    }
  }

  for (iBrd = 0; iBrd < g_demoEnv.coBrds; iBrd++) {

    // Connect (get handles) to the physical devices on the Thor-2-PCI board
    DemoLogf("Opening physical devices...\n");
    DemoOpenPhysicalDevices(iBrd);

  }

  // Inform the driver that our two ThorController boards are connected to one CT Bus
  for (iBrd = 0; iBrd < g_demoEnv.coBrds; iBrd++) {
    result = OtxDrvConnectBoard(g_hCtBus[iBrd/2], g_hPhysDevs[iBrd].hThorController); DemoCheckRc(g_hCtBus[iBrd/2], result);
  }

    
  // Set clocks sources for the boards
  {
    OTX_INT32 nCtBus;

    for (nCtBus = 0; nCtBus < g_demoEnv.coCtBuses; nCtBus++) {
      OTX_INT32 nBrdA = nCtBus*2+0;
      OTX_INT32 nBrdB = nCtBus*2+1;

      result = OtxBrdSetClocks(
        g_hPhysDevs[nBrdB].hThorController, 
        g_demoEnv.eClkMode, 
        g_demoEnv.eClkSource); 
      DemoCheckRc(g_hPhysDevs[nBrdB].hThorController, result);

      result = OtxBrdSetClocks(
        g_hPhysDevs[nBrdA].hThorController, 
        g_demoEnv.eClkMode, 
        g_demoEnv.eClkMode == OTX_CLOCK_MODE_H100_ETCF ? OTX_CLOCK_SOURCE_A_CLOCKS_ETCF : OTX_CLOCK_SOURCE_A_CLOCKS); 
      DemoCheckRc(g_hPhysDevs[nBrdA].hThorController, result);
    }
  }

  for (iBrd = 0; iBrd < g_demoEnv.coBrds; iBrd++) {
    //Configuring E1/T1 devices
    DemoConfigureE1T1Devices(iBrd);

    // Create and get handles to the Logical devices needed by this application
    DemoCreateLogicalDevices(iBrd);
    
    // Create and get handles to the Pipe devices needed by this application
    DemoConnectDevices(iBrd);
    
    // Initialized the Connected Logical Devices
    DemoInitializeLogicalDevices(iBrd);

    // Enable all devices. Their device state should now be OTX_DEVICE_WORKING
    DemoEnableDevices(iBrd);
    
  }


  // Read Serial Number from the boards
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;
    char *pszBrdBuildNumber;

    for (iBrd = 0; iBrd < g_demoEnv.coBrds; iBrd++) {
      result = OtxDrvGetAttributeType(g_hPhysDevs[iBrd].hThorController, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs[iBrd].hThorController, result);
      assert(eAttribType == OTX_ATTRIB_SZ);
      assert(bReadOnly == OTX_TRUE);
      pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
      result = OtxDrvGetAttributeValueSz(g_hPhysDevs[iBrd].hThorController, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs[iBrd].hThorController, result);
      DemoLogf("Board#%d Serial Number: %s\n", iBrd, pszSerialNumber);
      free(pszSerialNumber);

      result = OtxDrvGetAttributeType(g_hPhysDevs[iBrd].hThorController, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs[iBrd].hThorController, result);
      assert(eAttribType == OTX_ATTRIB_SZ);
      assert(bReadOnly == OTX_TRUE);
      pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
      result = OtxDrvGetAttributeValueSz(g_hPhysDevs[iBrd].hThorController, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs[iBrd].hThorController, result);
      DemoLogf("Board#%d: Board Build Number: %s\n", iBrd, pszBrdBuildNumber);
      free(pszBrdBuildNumber);
      
    }
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

	OsaThreadSleep(300);
  DemoLogf("\nPress 'h' for help\n");
  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

#if defined(_WIN32_WCE)
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd) 
{
  // default command line parameters
  static char exename[OTX_MAX_PATH * sizeof(TCHAR)] = "";
  static char *cmdParams[] = { exename, "/ME", NULL }; 

  GetModuleFileName(NULL, (LPTSTR) cmdParams[0], OTX_MAX_PATH);
  TCHARTOSTR((LPTSTR) cmdParams[0], cmdParams[0]);

  return OsaWinCeConsoleMain(hInstance, hPrevInstance, lpCmdLine, nShowCmd,
              cmdParams, sizeof(cmdParams) / sizeof(cmdParams[0]), main);
}
#endif  // #if defined(_WIN32_WCE)
